-- Discriminated Unions can be used to create complex
-- data types easily. This is a simple case, similar
-- to an enum in .NET:

data Color = R | G | B
	deriving (Eq, Show)

-- The elements following the = and separated by | 
-- are called data constructors. This becomes clearer
-- with an example where there are data elements:

type Text = String
type Checked = Bool

data UIElement = GroupBox Text [UIElement] 
			   | Button Text
			   | Label Text
			   | Edit Text
			   | CheckBox Text Checked
			   deriving (Show)
				
createTestUI = 
	GroupBox "Some Controls" [
		Label "Please enter your name here",
		Edit "<name>",
		CheckBox "Remember me" True,
		Button "OK", Button "Cancel"
	]

createLine n = take n (repeat '-')

formatElements :: [UIElement] -> String
formatElements [] = []
formatElements (el:els) = formatElement el ++ formatElements els

formatElement :: UIElement -> String
formatElement (GroupBox t els) =
	"--- " ++ t ++ " ---------------\n" ++
	formatElements els ++ createLine ((length t) + 20)

formatElement (Label t) = t ++ "\n\n"
formatElement (Edit t) = "[ " ++ t ++ " ]\n\n"
formatElement (CheckBox t c) = "[" ++ (if c then "X" else "")  ++ "] " ++ t ++ "\n\n"
formatElement (Button t) = 
	let w = (length t) + 4
	in createLine w ++ "\n| " ++ t ++ " |\n" ++ createLine w ++ "\n"

outputElements :: [UIElement] -> IO()
outputElemests [] = do return
outputElements l@(x:xs) = sequence_ [putStrLn (formatElement el) | el <- l]

-- Types can also be "generic":

data List a = Cons a (List a)
			      | Empty
			        deriving (Show)

-- A few simple conversion functions - they also show the relationship to standard lists
fromList (x:xs) = Cons x (fromList xs)
fromList [] = Empty

fromCons (Cons x xs) = x : (fromCons xs)
fromCons Empty = []

