-- Maybe I have my own data type

data Food = Pasta | Pizza | Chips

-- I might have a function to check for equality like this:

isEqualFood :: Food -> Food -> Bool
isEqualFood Pasta Pasta = True
isEqualFood Pizza Pizza = True
isEqualFood Chips Chips = True
isEqualFood _ _ = False

-- Main> isEqualFood Pizza Chips
-- False
-- Main> isEqualFood Pizza Pizza
-- True
-- Main> 

-- Checking for equality is a standard thing though, something 
-- I might want to do for other types as well. A type class defines
-- a class of types that can do something, in this case equality 
-- comparisons.
-- The type class *can* provide default implementations of functions.
-- In this case both functions have a default implementation, so when
-- an instance is created, the programmer has a choice of which 
-- implementation to provide. Of course, she has to provide at least
-- one of the implementations to resolve the circular dependency.

class MyEq a where
	isEqual :: a -> a -> Bool
	isEqual x y = not (isNotEqual x y)
	
	isNotEqual :: a -> a -> Bool
	isNotEqual x y = not (isEqual x y)

-- The type class makes its functions available
-- 
-- Main> :type isEqual
-- isEqual :: MyEq a => a -> a -> Bool
-- Main> 

-- I must generate instances of the type class for specific
-- types in order to say "this type is now among those that
-- can do what the type class requires", i.e. in this case
-- perform comparisons.

instance MyEq Food where
	isEqual Pasta Pasta = True
	isEqual Pizza Pizza = True
	isEqual Chips Chips = True
	isEqual _ _ = False

-- I can now compare values of type Food via the MyEq type class:
-- 
-- Main> isEqual Pizza Chips
-- False
-- Main> isEqual Pizza Pizza
-- True
-- Main> 
	
-- Type class instances can also be created for existing types:
	
instance MyEq Bool where 
	isEqual True True = True
	isEqual False False = True
	isEqual _ _ = False

-- Main> isEqual True True
-- True
-- Main> isEqual True False
-- False
-- Main> 

-- One thing I can't do right now is print out a value of type Food:

-- Main> Chips
-- ERROR - Cannot find "show" function for:
-- *** Expression : Chips
-- *** Of type    : Food
-- 
-- Main> 

-- The Haskell Prelude comes with a number of standard type classes.
-- Show is used to convert a value into a string (the opposite, Read, 
-- exists as well).
	
instance Show Food where
	show Chips = "Chips"
	show Pizza = "Pizza"
	show Pasta = "Pasta"

-- Now, Food types can output themselves because they are valid arguments
-- for the show function:

-- Main> Chips
-- Chips
-- Main> show Chips
-- "Chips"
-- Main> 	
	
-- For the four standard type classes Read, Show, Eq and Ord, the Haskell
-- Report requires the ability of the compiler to auto-generate an 
-- implementation. 
	
data Food2 = Fish | Steak | Noodles
	deriving (Read, Show, Eq, Ord)

-- This allows Food2 to output itself, but also to support the standard
-- equality operators == and /=, which Food still doesn't support

-- Main> Fish
-- Fish
-- Main> Fish == Noodles
-- False
-- Main> Fish == Fish
-- True
-- Main> Chips == Chips
-- ERROR - Cannot infer instance
-- *** Instance   : Eq Food
-- *** Expression : Chips == Chips
-- 
-- Main> 
	
-- Type inference in Haskell takes advantage of type classes.
-- This function adds two values:
	
add x y = x + y

-- Since the operator (+) is defined as a function in the type
-- class Num, the type of add is inferred as such:

-- Main> :type add
-- add :: Num a => a -> a -> a
-- Main> 

-- Different types have instances for Num, which allows me to use
-- the add function with Integers and floating point numbers - 
-- this does *not* work in F#, for instance:

-- Main> add 10 20
-- 30
-- Main> add 3.5 7.4
-- 10.9
-- Main> 

-- I can create instances of standard type classes for my own types:

instance Num Food2 where
	(+) Fish Noodles = Steak
	(+) Noodles Noodles = Fish

-- This enables me to add certain elements of Food2 now:

-- Main> Fish + Noodles
-- Steak
-- Main> Noodles + Noodles
-- Fish
-- Main> Steak + Fish
-- 
-- Program error: pattern match failure: v1621_v1665 Steak Fish
-- 
-- Main> 